#include <endian.h>
#include <sys/stat.h>
#include <errno.h>

#include "store_mdb.h"


#if __BYTE_ORDER == __BIG_ENDIAN  // FIXME not portable.
  #define REVERSEKEY 0
  #define REVERSEDUP 0
#else
  #define REVERSEKEY MDB_REVERSEKEY
  #define REVERSEDUP MDB_REVERSEDUP
#endif

#define INT_KEY_MASK MDB_INTEGERKEY | REVERSEKEY
#define INT_DUP_KEY_MASK MDB_DUPSORT | MDB_DUPFIXED \
    | MDB_INTEGERKEY | REVERSEKEY | REVERSEDUP

#define INT_DUP_MASK MDB_DUPSORT | MDB_DUPFIXED \
    | MDB_INTEGERDUP | REVERSEKEY | REVERSEDUP

/**
 * Number of DBs defined.
 */
#define N_DB 13

#define ENV_DIR_MODE 0750
#define ENV_FILE_MODE 0640


typedef char DbLabel[8];


/**
 * Static handles.
 */
static char *env_path = NULL;
static bool env_init = false;
static bool db_init = false;


/**
 * Main DBs. These are the master information containers.
 *
 * Data columns are: identifier prefix, DB label, flags.
 */
#define MAIN_TABLE \
    ENTRY(  T_ST,    "t:st",    INT_KEY_MASK    )   /* Key to ser. term */  \
    ENTRY(  SPO_C,   "spo:c",   INT_DUP_MASK    )   /* Triple to context */ \
    ENTRY(  C_,      "c:",      INT_KEY_MASK    )   /* Track empty contexts */\
    ENTRY(  PFX_NS,  "pfx:ns",  0               )   /* Prefix to NS */      \

/**
 * Lookup DBs. These are indices and may be destroyed and rebuilt.
 */
#define LOOKUP_TABLE \
    ENTRY(  S_PO,    "s:po",    INT_DUP_KEY_MASK)   /* 1-bound lookup */    \
    ENTRY(  P_SO,    "p:so",    INT_DUP_KEY_MASK)   /* 1-bound lookup */    \
    ENTRY(  O_SP,    "o:sp",    INT_DUP_KEY_MASK)   /* 1-bound lookup */    \
    ENTRY(  PO_S,    "po:s",    INT_DUP_MASK    )   /* 2-bound lookup */    \
    ENTRY(  SO_P,    "so:p",    INT_DUP_MASK    )   /* 2-bound lookup */    \
    ENTRY(  SP_O,    "sp:o",    INT_DUP_MASK    )   /* 2-bound lookup */    \
    ENTRY(  C_SPO,   "c:spo",   INT_DUP_KEY_MASK)   /* Context lookup */    \
    ENTRY(  TH_T,    "th:t",    0               )   /* Term hash to key */  \
    ENTRY(  NS_PFX,  "ns:pfx",  0               )   /* NS to prefix */      \

/**
 * DB labels. They are prefixed with DB_
 */
#define ENTRY(a, b, c) static const DbLabel DB_##a = b;
MAIN_TABLE
LOOKUP_TABLE
#undef ENTRY

/**
 * Numveric index of each DB in the following constants. Prefixed with IDX_
 */
typedef enum {
#define ENTRY(a, b, c) IDX_##a,
    MAIN_TABLE
    LOOKUP_TABLE
#undef ENTRY
} DBIdx;

/**
 * DB indices. These are populated on init.
 */
static MDB_dbi dbis[N_DB];

/**
 * DB labels.
 */
static const char *db_labels[N_DB] = {
#define ENTRY(a, b, c) DB_##a,
    MAIN_TABLE
    LOOKUP_TABLE
#undef ENTRY
};

/**
 * DB flags. These are aligned with the dbi_labels index.
 */
static const unsigned int db_flags[N_DB] = {
#define ENTRY(a, b, c) c,
    MAIN_TABLE
    LOOKUP_TABLE
#undef ENTRY
};

/**
 * 1-bound and 2-bound lookup indices.
 *
 * N.B. Only the first 6 (1-bound and 2-bound term lookup) are used.
 * The others are added just because they belong logically to the lookup table.
 */
static DBIdx lookup_indices[9] = {
#define ENTRY(a, b, c) IDX_##a,
    LOOKUP_TABLE
#undef ENTRY
};

/**
 * Order in which keys are looked up if two terms are bound.
 * The indices with the smallest average number of values per key should be
 * looked up first.
 *
 * 0 = s:po
 * 1 = p:so
 * 2 = o:sp
 */
static const uint8_t lookup_rank[3] = {0, 2, 1};

static const uint8_t lookup_ordering_1bound[3][3] = {
    {0, 1, 2}, // spo
    {1, 0, 2}, // pso
    {2, 0, 1}, // osp
};

static const uint8_t lookup_ordering_2bound[3][3] = {
    {1, 2, 0}, // po:s
    {0, 2, 1}, // so:p
    {0, 1, 2}, // sp:o
};


/**
 * Static prototypes.
 */
static int _dbi_init(bool create);
static int _mdbenv_init(bool create);
static void _env_cleanup();


/**
 * API.
 */
int
LSUP_store_open(bool create)
{
    if(UNLIKELY(!env_init)) _mdbenv_init(create);

    atexit(_env_cleanup);

    int rc = mdb_env_open(LSUP_mdbenv, env_path, 0, ENV_FILE_MODE);

    if (UNLIKELY(!db_init)) rc |= _dbi_init(create);

    return rc;
}


int
LSUP_store_stats()
{
    // TODO
    MDB_stat env_stat, db_stats[N_DB];
    return 0;
}


size_t
LSUP_store_size(MDB_txn *txn)
{
    if (!env_init) return 0;

    MDB_stat stat;
    mdb_stat(txn, dbis[IDX_SPO_C], &stat);

    return stat.ms_entries;
}


/**
 * Static functions.
 */

/**
 * @brief Create and initialize the MDB env.
 *
 * This function takes care of creaating the environment path if not existing,
 * and checking that it's a writable directory. If the path is not specified
 * in the LSUP_STORE_PATH environment variable, a temporary file is used.
 *
 * This should be only called once per program run, therefore it's best to let
 * LSUP_mdb_env_open() call it when needed.
 *
 * TODO create is not used; either use it or get rid of it.
 */
static int
_mdbenv_init(bool create)
{
    env_path = getenv("LSUP_STORE_PATH");
    struct stat path_stat;
    int rc;

    // Set environment path.
    if (env_path == NULL) {
        printf(
                "WARNING: `LSUP_STORE_PATH' environment variable is not set. "
                "A temporary store path will be open and will be DESTROYED "
                "at the closing of the program.");
        env_path = mkdtemp("lsup_mdb-XXXXXX");
    } else {

        rc = stat(env_path, &path_stat);
        if (rc == ENOENT) {
            if (mkdir(env_path, ENV_DIR_MODE) != 0) abort();
        } else {
            if (!S_ISDIR(path_stat.st_mode)) abort();
        }
    }

    mdb_env_create(&LSUP_mdbenv);

    // Set map size.
    size_t mapsize;
    char *env_mapsize = getenv("LSUP_MDB_MAPSIZE");

    if (env_mapsize == NULL) {
        mapsize = 1024LU << 30; // 1Tb
    } else {
        sscanf(env_mapsize, "%lu", &mapsize);
    }

    mdb_env_set_mapsize(LSUP_mdbenv, mapsize);

    mdb_env_set_maxdbs(LSUP_mdbenv, N_DB);

    env_init = true;

    return rc;

}


/**
 * @brief Open and allocate DB handles in an array.
 *
 * @param bool create [in]: If true, the DBs are created. This is only needed
 *  on bootstrap.
 */
static int
_dbi_init(bool create)
{
    MDB_txn *txn;
    unsigned int txn_flags = create ? 0 : MDB_RDONLY;
    unsigned int create_flag = create ? MDB_CREATE : 0;

    mdb_txn_begin(LSUP_mdbenv, NULL, txn_flags, &txn);

    for (int i = 0; i < N_DB; i++) {
        mdb_dbi_open(txn, db_labels[i], db_flags[i] | create_flag, dbis + i);
    }

    mdb_txn_commit(txn);

    return 0;
}


/* TODO */
static void
_env_cleanup()
{
    if (env_init) {
        printf("Cleaning up MDB env.\n");
        mdb_env_close(LSUP_mdbenv);
    }
}

